# Generated by Django 4.2 on 2023-05-10 11:08

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("permissions", "0004_create_anonymous_user"),
        ("observers", "0003_alter_observer_object_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="backgroundtask",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(class)s",
                to="permissions.permissiongroup",
            ),
        ),
    ]



""".. Ignore pydocstyle D400.

===================
Migrate annotations
===================

Migrate annotations from old schema to the new one. All new annotations will be
re-created from old ones.
"""

import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterable, Set, Union

import openpyxl
from django.core.management.base import BaseCommand
from django.db import models, transaction

from resolwe.flow.models import (
    AnnotationField,
    AnnotationGroup,
    AnnotationValue,
    Collection,
    DescriptorSchema,
    Entity,
)
from resolwe.flow.models.annotations import AnnotationType
from resolwe.flow.utils import iterate_schema


def read_mapping(file: Union[str, os.PathLike]):
    """Read the mappings file."""

    def get_row(excel_row, header):
        """Get the dictionary with header values for keys."""
        return {
            column_name: excel_row[column_index].value
            for column_index, column_name in header.items()
        }

    workbook = openpyxl.load_workbook(file)
    sheet = workbook["CHANGELOG"]

    type_map = {
        "basic:string": AnnotationType.STRING,
        "basic:decimal": AnnotationType.DECIMAL,
        "basic:integer": AnnotationType.INTEGER,
        "basic:date": AnnotationType.DATE,
    }
    header = {
        index: cell.value
        for index, cell in enumerate(next(sheet.rows))
        if cell.value is not None
    }

    # Read all the group labels in advance.
    group_labels = dict()
    for excel_row in sheet.iter_rows(2):
        row = get_row(excel_row, header)
        if row["GROUP/FIELD"] == "GROUP" and row["FROM"].strip() != "-":
            group_labels[row["NEW group name"]] = row["Label"]

    mapping = dict()
    for excel_row in sheet.iter_rows(2):
        row = get_row(excel_row, header)
        if row["GROUP/FIELD"] != "FIELD" or row["TO"].lower().strip() == "obsolete":
            continue

        choice_row = row["Choices {label : value}"]
        choices = choice_row.strip(" {}").split(", ") if choice_row else []
        vocabulary = {
            choice.split(":")[1].strip(): choice.split(":")[0].strip()
            for choice in choices
        }

        for processing_group in row["OLD group name"].split(" or "):
            field = f"{processing_group}.{row['FROM']}"
            mapping[field] = (
                row["NEW group name"],
                group_labels[row["NEW group name"]],
                row["TO"],
                row["Label"],
                row["Description"],
                type_map[row["Type"]],
                vocabulary,
            )
    return mapping


def process_descriptor_schema(
    descriptor_schema: Dict, mapping: Dict, field_map: Dict
) -> Iterable[AnnotationField]:
    """Get the annotation fields from the descriptor schema."""
    fields = set()
    for schema, _, field_full_path in iterate_schema({}, descriptor_schema):
        # Field is a dictionary contating keys name, label, type and required.
        assert field_full_path in mapping, f"{field_full_path} is not known"
        (
            group_name,
            group_label,
            field_name,
            field_label,
            field_description,
            field_type,
            field_vocabulary,
        ) = mapping[field_full_path]
        # TODO: group sort order.
        group = AnnotationGroup.objects.get_or_create(
            name=group_name, label=group_label
        )[1]
        # TODO: field sort order, field description.
        field = AnnotationField.objects.get_or_create(
            name=field_name,
            type=field_type,
            group=group,
            vocabulaly=field_vocabulary,
            label=field_label,
            description=field_description,
            validator_regex=schema.get("validate_regex"),
        )[1]
        field_map[field_full_path] = field
        fields.add(field)
    return fields


@transaction.atomic
def migrate_descriptors(mapping_file: str):
    """Create new descriptors from old ones."""
    # This should delete all the new annotation objects. Hopefully it will be fast enough.
    AnnotationGroup.objects.all().delete()
    mapping_path = Path(mapping_file)
    mapping = read_mapping(mapping_path)
    # Iterate through all descriptor schemas that are set on the Collections.
    used_descriptor_schemas = DescriptorSchema.objects.filter(
        models.Q(collection__isnull=False) | models.Q(entity__isnull=False)
    ).distinct()

    visited_fields = dict()
    for descriptor_schema in used_descriptor_schemas.iterator():
        for field, _, field_full_path in iterate_schema({}, descriptor_schema.schema):
            # Field is a dictionary contating keys name, label, type and required.
            visited_fields[field_full_path] = field

    missing_fields = set()
    for field_path in visited_fields:
        if field_path not in mapping:
            missing_fields.add(field_path)

    # Mapping between full field name and field instance.
    field_map = dict()
    # Mapping between collection and fields assigned to it.
    collection_annotation_fields: Dict[int, Set[AnnotationField]] = defaultdict(set)
    # Mapping from descriptor schema id to the list of annotation field.
    schema_fields = dict()
    # Now iterate through all the collections. Read the annotation schema of
    # its entities and create corresponding fields.
    for collection_id, schema_id, schema in (
        Entity.objects.filter(descriptor_schema__isnull=False, collection__isnull=False)
        .values_list("collection", "descriptor_schema", "descriptor_schema__schema")
        .iterator()
    ):
        if schema_id not in schema_fields:
            schema_fields[schema_id] = process_descriptor_schema(
                schema, mapping, field_map
            )
        collection_annotation_fields[collection_id].update(schema_fields[schema_id])

    for collection_id, collection_fields in collection_annotation_fields.values():
        collection: Collection = Collection.objects.get(pk=collection_id)
        collection.annotation_fields.set(collection_fields)
        # Process values for all entities in the collection.
        for entity in collection.entity_set.all():
            for _, field_value, field_full_path in iterate_schema(
                entity.descriptor, entity.descriptor_schema.schema
            ):
                field = field_map[field_full_path]
                AnnotationValue.objects.get_or_create(
                    entity=entity, field=field, value=field_value
                )


class Command(BaseCommand):
    """Migrate descriptors."""

    help = "Migrate descriptors."

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "mapping_file",
            help="The file with mappings from the old annotation schema to fields.",
        )

    def handle(self, *args, **options):
        """Migrate descriptors."""
        migrate_descriptors(options["mapping_file"])



# Generated by Django 4.2.11 on 2024-03-18 13:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0022_fix_group_by"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="entity",
            name="descriptor",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="descriptor_dirty",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="descriptor_schema",
        ),
        migrations.RemoveField(
            model_name="process",
            name="entity_descriptor_schema",
        ),
    ]



"""Runtimes manager."""

import logging
from contextlib import suppress
from importlib import import_module
from types import ModuleType
from typing import Dict, Optional, Set, Type

from .runtime import Process

logger = logging.getLogger(__name__)


class RegisteredRuntimes:
    """A list of registered base python processes to be copied into executor.

    The python processes is must extend the base class for
    python processes :class:`resolwe.process.runtime.Process` and set the
    class attribute `_abstract` to True. They must also be registered by
    addind their name to the settings `FLOW_PROCESSES_RUNTIMES`.

    Is such case the entire parent module of the file in which the process is
    defined is copied to the executor. It is the responsibility of the
    programer to make sure that the module uses only Python standard library
    (and possibly other runtimes it inherits from).
    """

    def __init__(self):
        """Initialize."""
        self._class_module: Dict[str, ModuleType] = dict()
        self._registered_classes: Dict[str, Type[Process]] = dict()
        # A set of dependencies for every registered class.
        self._dependencies: Dict[str, Set[ModuleType]] = dict()
        self._register_runtimes()

    def _register_runtimes(self):
        """Register python runtimes."""
        with suppress(ImportError):
            from django.conf import settings

            for runtime_class_name in getattr(
                settings,
                "FLOW_PROCESSES_RUNTIMES",
                ("resolwe.process.runtime.Process",),
            ):
                module_name, class_name = runtime_class_name.rsplit(".", 1)
                self.register_runtime(getattr(import_module(module_name), class_name))

    def _get_runtime_classes(self, klass: Type["Process"]) -> Set[Type["Process"]]:
        """Get all the runtime classes of a given class.

        This includes all the runtime classes this class inherits from.
        """
        bases = set()
        if "_abstract" in klass.__dict__ and issubclass(klass, Process):
            bases.add(klass)
        for base_class in klass.__bases__:
            bases.update(self._get_runtime_classes(base_class))
        return bases

    def register_runtime(self, klass: Type["Process"]):
        """Register base class for processes.

        :raises AssertionError: if runtime is already registered or klass is
            not a base class for python processes.
        """
        class_name = klass.__name__
        assert (
            class_name not in self._registered_classes
        ), f"Class {class_name} already registered."
        assert issubclass(
            klass, Process
        ), f"Class {class_name} must be a subclass of resolwe.process.runtime.Process."
        assert (
            "_abstract" in klass.__dict__
        ), f"Class {class_name} must contain as class attribute _abstract."

        module_name = klass.__module__[: klass.__module__.rindex(".")]
        module = import_module(module_name)

        self._registered_classes[klass.__name__] = klass
        self._class_module[klass.__name__] = module
        self._dependencies[klass.__name__] = {
            self._class_module[runtime_class.__name__]
            for runtime_class in self._get_runtime_classes(klass)
        }
        logger.debug(
            "Registered python class %s, module %s.", klass.__name__, module.__name__
        )

    def necessary_modules(self, class_name: str) -> Set[ModuleType]:
        """Get a set of necessary modules for the given class_name."""
        return self._dependencies[class_name]

    def registered_class(self, class_name: str) -> Optional[Type["Process"]]:
        """Get registered class from name."""
        return self._registered_classes.get(class_name)

    def is_registered(self, class_name: str) -> bool:
        """Return if the given class name is registered as runtime class."""
        return class_name in self._registered_classes


python_runtimes_manager = RegisteredRuntimes()



# Generated by Django 4.2.9 on 2024-02-13 07:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0025_add_unaccent_extension"),
    ]

    operations = [
        migrations.RunSQL(
            """
            DO
            $$BEGIN
                CREATE TEXT SEARCH CONFIGURATION simple_unaccent( COPY = simple );
            EXCEPTION
                WHEN unique_violation THEN
                    NULL;  -- ignore error
            END;$$;
            """
        ),
        migrations.RunSQL(
            "ALTER TEXT SEARCH CONFIGURATION simple_unaccent "
            + "ALTER MAPPING FOR hword, hword_part, word "
            + "WITH unaccent, simple;"
        ),
    ]



""".. Ignore pydocstyle D400.

======================
Flow Executor Preparer
======================

Framework for the manager-resident executor preparation facilities.

.. autoclass:: resolwe.flow.executors.prepare.BaseFlowExecutorPreparer
    :members:

"""

import logging
import os

from django.conf import settings

from resolwe.flow.utils import get_apps_tools
from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors
from resolwe.test.utils import is_testing

logger = logging.getLogger(__name__)


class BaseFlowExecutorPreparer:
    """Represents the preparation functionality of the executor."""

    def prepare_for_execution(self, data):
        """Prepare the data object for the execution.

        This is mostly needed for the null executor to change the status of
        the data and worker object to done.
        """

    def extend_settings(self, data_id, files, secrets):
        """Extend the settings the manager will serialize.

        :param data_id: The :class:`~resolwe.flow.models.Data` object id
            being prepared for.
        :param files: The settings dictionary to be serialized. Keys are
            filenames, values are the objects that will be serialized
            into those files. Standard filenames are listed in
            ``resolwe.flow.managers.protocol.ExecutorFiles``.
        :param secrets: Secret files dictionary describing additional secret
            file content that should be created and made available to
            processes with special permissions. Keys are filenames, values
            are the raw strings that should be written into those files.
        """
        pass

    def get_tools_paths(self, from_applications=False):
        """Get tools' paths."""
        if settings.DEBUG or is_testing() or from_applications:
            return list(get_apps_tools().values())

        else:
            tools_root = storage_settings.FLOW_VOLUMES["tools"]["config"]["path"]
        try:
            subdirs = next(os.walk(tools_root))[1]
            return [os.path.join(tools_root, sdir) for sdir in subdirs]
        except StopIteration:
            return []

    def post_register_hook(self, verbosity=1):
        """Run hook after the 'register' management command finishes.

        Subclasses may implement this hook to e.g. pull Docker images at
        this point. By default, it does nothing.
        """

    def resolve_data_path(self, data=None, filename=None):
        """Resolve data path for use with the executor.

        :param data: Data object instance
        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given data file in programs executed using this executor
        """
        if data is None:
            filesystem_connectors = [
                connector
                for connector in connectors.for_storage("data")
                if connector.mountable
            ]
            return filesystem_connectors[0].path if filesystem_connectors else None

        return data.location.get_path(filename=filename)

    def resolve_upload_path(self, filename=None):
        """Resolve upload path for use with the executor.

        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given uploaded file in programs executed using this
            executor

        :raises RuntimeError: when no storage connectors are configured for
          upload storage or path could not be resolved.
        """
        upload_connectors = [
            connector
            for connector in connectors.for_storage("upload")
            if connector.mountable
        ]
        if not upload_connectors:
            raise RuntimeError("No connectors are configured for 'upload' storage.")

        upload_connector = upload_connectors[0]
        if filename is None:
            return upload_connector.path
        else:
            return os.path.join(upload_connector.path, filename)

    def get_environment_variables(self):
        """Return dict of environment variables that will be added to executor."""

        return {}



""".. Ignore pydocstyle D400.

====================
Resolwe Test Helpers
====================

"""

TESTING_CONTEXT = {
    "is_testing": False,
}


def is_testing():
    """Return current testing status."""
    return TESTING_CONTEXT["is_testing"]



# Generated by Django 3.2.16 on 2023-03-31 08:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("observers", "0002_backgroundtask"),
    ]

    operations = [
        migrations.AlterField(
            model_name="observer",
            name="object_id",
            field=models.PositiveIntegerField(default=0),
            preserve_default=False,
        ),
    ]



"""Report the listener metrics to the monitoring system."""

import logging
import os
import socket
from contextlib import suppress
from time import time
from typing import Any

from opentelemetry import metrics
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource

from resolwe.flow.executors.socket_utils import (
    Message,
    MessageProcessingCallback,
    MessageProcessingEventType,
    PeerIdentity,
)

logger = logging.getLogger(__name__)

# How often to export metrics (in milliseconds).
METRICS_EXPORT_INTERVAL = 30000
# TODO: Remove this when metrics are properly implemented.
METRICS_INSECURE_EXPORT = True
# The number of data points to send in one request.
# See
# https://github.com/open-telemetry/opentelemetry-python/pull/2809
# and
# https://github.com/open-telemetry/opentelemetry-python/issues/2710
# for details.
try:
    METRICS_EXPORT_SIZE = int(os.environ.get("METRICS_EXPORT_SIZE", 1000))
except ValueError:
    METRICS_EXPORT_SIZE = 1000


class MetricsEventReporter(MessageProcessingCallback):
    """Class handling the reporting of the metrics events."""

    # The prefix used in all metric names. The name is generated as
    # f"{self.metrics_prefix}_{metric_name}", so avoid adding '_' at the end.
    metrics_prefix = "resolwe_listener"

    def __init__(self):
        """Initialize the metrics event reporter."""
        if (endpoint := os.environ.get("METRICS_ENDPOINT")) is None:
            logger.info("Metrics endpoint is not set, reporting disabled.")
            self._enabled = False
            logger.warning(
                "Listener metrics endpoint is not set, diabling metrics reporting."
            )
        else:
            logger.info("Reporting metrics to endpoint %s.", endpoint)
            self._hostname = socket.gethostname()
            self._enabled = True
            self._init_metrics(endpoint)

    def _init_metrics(self, metric_endpoint):
        # Initialize the counters for the metrics.
        self._processing_messages = 0
        self._queued_messages = 0

        # Create the dictionaries with the appropriate timestamps.
        self._processing_started: dict[bytes, float] = {}
        self._message_received: dict[bytes, float] = {}

        # Initialize the opentelemetry metrics.
        metric_exporter = OTLPMetricExporter(
            endpoint=metric_endpoint,
            insecure=METRICS_INSECURE_EXPORT,
            max_export_batch_size=METRICS_EXPORT_SIZE,
        )
        metric_reader = PeriodicExportingMetricReader(
            metric_exporter, METRICS_EXPORT_INTERVAL
        )
        provider = MeterProvider(
            metric_readers=[metric_reader],
            resource=Resource.create({"service.name": "listener"}),
        )

        # Sets the global default meter provider
        metrics.set_meter_provider(provider)

        # Creates a meter from the global meter provider
        meter = metrics.get_meter("listener")

        # Create the metrics.
        self._messages_queued = meter.create_up_down_counter(
            name=f"{self.metrics_prefix}_messages_queued",
            description="Number of unprocessed messages in the queue.",
        )
        self._messages_processing = meter.create_up_down_counter(
            name=f"{self.metrics_prefix}_messages_processing",
            description="Number of currently processing messages.",
        )
        self._preparation_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_preparation_duration",
            description="The duration of init phase.",
            unit="seconds",
        )
        self._response_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_response_duration",
            unit="miliseconds",
            description="Response time for messages.",
        )
        self._processing_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_processing_duration",
            unit="miliseconds",
            description="Processing time for messages.",
        )
        self._network_time_histogram = meter.create_histogram(
            name=f"{self.metrics_prefix}_network_duration",
            unit="miliseconds",
            description="Time messages spent in the transfer.",
        )

    def _counter_changed(
        self, counter: metrics.UpDownCounter, value: int, attributes: dict
    ):
        """Change the value of the given counter."""
        counter.add(value, attributes)

    def _attributes_from_message(self, message: Message) -> dict[str, Any]:
        """Create the attributes from the given message."""
        return {"command": message.type_data}

    def event(
        self,
        event_type: MessageProcessingEventType,
        message: Message,
        peer_identity: PeerIdentity,
        **kwargs,
    ):
        """Process the event from the message processing pipeline."""
        # Do not proceed if metric reporting is disabled.
        if not self._enabled:
            return

        if message.client_id is None:
            logger.error("Message client_id is not defined.")
            return

        attributes = self._attributes_from_message(message)
        attributes["pod"] = self._hostname
        with suppress(Exception):
            attributes["data_id"] = abs(int(peer_identity))

        match event_type:
            case MessageProcessingEventType.MESSAGE_RECEIVED:
                self._queued_messages += 1
                self._counter_changed(self._messages_queued, 1, attributes)
                self._message_received[message.client_id] = time()
                value = (
                    self._message_received[message.client_id] - message.sent_timestamp
                )
                self._network_time_histogram.record(value * 1000, attributes)

            case MessageProcessingEventType.MESSAGE_PROCESSING_STARTED:
                self._processing_messages += 1
                self._counter_changed(self._messages_processing, 1, attributes)
                self._processing_started[message.client_id] = time()

            case MessageProcessingEventType.PREPARATION_FINISHED:
                # This type expects the 'started' unix timestamp in kwargs.
                if "started" not in kwargs:
                    logger.error(
                        "Extra data 'started' expected for preparation finished event."
                    )
                    return
                preparation_time = time() - kwargs["started"]
                self._preparation_time_histogram.record(preparation_time, attributes)

            case MessageProcessingEventType.MESSAGE_PROCESSING_FINISHED:
                self._processing_messages -= 1
                self._queued_messages -= 1
                self._counter_changed(self._messages_processing, -1, attributes)
                self._counter_changed(self._messages_queued, -1, attributes)

                if message.client_id not in self._processing_started:
                    logger.error(
                        "Message from client '%s' not found in the processing messages.",
                        message.client_id,
                    )
                    return
                if message.client_id not in self._message_received:
                    logger.error(
                        "Message from client '%s' not found in the received messages.",
                        message.client_id,
                    )
                    return

                processing_started = self._processing_started.pop(message.client_id)
                message_received = self._message_received.pop(message.client_id)
                processing_finished = time()
                response_time = processing_finished - message_received
                processing_time = processing_finished - processing_started
                self._response_time_histogram.record(response_time * 1000, attributes)
                self._processing_time_histogram.record(
                    processing_time * 1000, attributes
                )


metrics_reporter = MetricsEventReporter()



"""Annotations viewset."""

from typing import Any

from rest_framework import (
    exceptions,
    generics,
    mixins,
    permissions,
    response,
    status,
    viewsets,
)
from rest_framework.serializers import BaseSerializer

from resolwe.flow.filters import (
    AnnotationFieldFilter,
    AnnotationPresetFilter,
    AnnotationValueFilter,
)
from resolwe.flow.models import AnnotationPreset
from resolwe.flow.models.annotations import AnnotationField, AnnotationValue
from resolwe.flow.serializers.annotations import (
    AnnotationFieldSerializer,
    AnnotationPresetSerializer,
    AnnotationValueSerializer,
)
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import Permission

from .mixins import ResolweCreateModelMixin, ResolweUpdateModelMixin


class AnnotationPresetViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`DescriptorSchema` objects."""

    serializer_class = AnnotationPresetSerializer
    permission_classes = (get_permissions_class(),)
    filterset_class = AnnotationPresetFilter

    # No need to specify default ordering: it is specified on the model.
    ordering_fields = ("id", "name", "sort_order")
    queryset = AnnotationPreset.objects.all()

    def get_queryset(self):
        """Get the presets for the current user."""
        return super().get_queryset().filter_for_user(self.request.user)


class AnnotationFieldViewSet(
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """Annotation fields viewset."""

    permission_classes = (permissions.AllowAny,)
    serializer_class = AnnotationFieldSerializer
    filterset_class = AnnotationFieldFilter
    # No need to specify default ordering: it is specified on the model.
    ordering_fields = ("id", "name", "label", "group__sort_order", "sort_order")
    queryset = AnnotationField.objects.all()


class AnnotationValueViewSet(
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.CreateModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
    generics.UpdateAPIView,
    viewsets.GenericViewSet,
):
    """Annotation value viewset."""

    serializer_class = AnnotationValueSerializer
    filterset_class = AnnotationValueFilter
    queryset = AnnotationValue.objects.all()
    permission_classes = (get_permissions_class(),)
    ordering_fields = ("modified", "id")

    def get_serializer(self, *args: Any, **kwargs: Any) -> BaseSerializer:
        """Get serializer instance depending on the request type."""
        kwargs_many = kwargs.get("many", False)
        kwargs["many"] = isinstance(self.request.data, list) or kwargs_many
        return super().get_serializer(*args, **kwargs)

    def _check_permissions(self, serializer):
        """Check if user has edit permission on entities."""
        validated_data = (
            serializer.validated_data
            if isinstance(serializer.validated_data, list)
            else [serializer.validated_data]
        )
        # Check permissions on entities.
        if not all(
            entity.has_permission(Permission.EDIT, self.request.user)
            for entity in {value["entity"] for value in validated_data}
        ):
            raise exceptions.PermissionDenied()

    def perform_create(self, serializer: BaseSerializer) -> None:
        """Perform create annotation value(s).

        The permission on entities must be checked.
        """
        self._check_permissions(serializer)
        return super().perform_create(serializer)

    def update(self, request, *args, pk=None, **kwargs):
        """Update annotation value(s).

        When posting multiple values, the request is treated as a bulk update. The bulk
        update can create, update or delete values. Values are deleted when the value
        is set no None.
        """
        # Regular update on a detail view.
        if pk is not None:
            return super().update(request, *args, pk=pk, **kwargs)

        # Bulk update / create / delete.
        serializer = self.get_serializer(data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        self._check_permissions(serializer)
        serializer.update(None, serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return response.Response(
            serializer.data, status=status.HTTP_200_OK, headers=headers
        )



"""Resolwe collection serializer."""

import logging
from typing import Optional

from rest_framework import serializers

from resolwe.flow.models import Collection, Data, DescriptorSchema
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .descriptor import DescriptorSchemaSerializer
from .fields import DictRelatedField

logger = logging.getLogger(__name__)


class BaseCollectionSerializer(ResolweBaseSerializer):
    """Base serializer for Collection objects."""

    settings = ProjectableJSONField(required=False)
    data_count = serializers.SerializerMethodField(required=False)
    status = serializers.SerializerMethodField(required=False)

    def get_data_count(self, collection: Collection) -> int:
        """Return number of data objects on the collection."""
        # Use 'data_count' attribute when available. It is created in the
        # BaseCollectionViewSet class.
        return (
            collection.data_count
            if hasattr(collection, "data_count")
            else collection.data.count()
        )

    def get_status(self, collection: Collection) -> Optional[str]:
        """Return status of the collection based on the status of data objects.

        When collection contains no data objects None is returned.
        """

        status_order = [
            Data.STATUS_ERROR,
            Data.STATUS_UPLOADING,
            Data.STATUS_PROCESSING,
            Data.STATUS_PREPARING,
            Data.STATUS_WAITING,
            Data.STATUS_RESOLVING,
            Data.STATUS_DONE,
        ]

        # Use 'data_statuses' attribute when available. It is created in the
        # BaseCollectionViewSet class. It contains all the distinct statuses of the
        # data objects in the collection.
        status_set = (
            set(collection.data_statuses)
            if hasattr(collection, "data_statuses")
            else collection.data.values_list("status", flat=True).distinct()
        )

        if not status_set:
            return None

        for status in status_order:
            if status in status_set:
                return status

        logger.warning(
            "Could not determine the status of a collection.",
            extra={"collection": collection.__dict__},
        )
        return None

    class Meta:
        """CollectionSerializer Meta options."""

        model = Collection
        read_only_fields = (
            "created",
            "duplicated",
            "id",
            "modified",
            "data_count",
            "status",
        )
        update_protected_fields = ("contributor",)
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "description",
                "name",
                "settings",
                "slug",
                "tags",
            )
        )


class CollectionSerializer(BaseCollectionSerializer):
    """Serializer for Collection objects."""

    descriptor = ProjectableJSONField(required=False)
    descriptor_schema = DictRelatedField(
        queryset=DescriptorSchema.objects.all(),
        serializer=DescriptorSchemaSerializer,
        allow_null=True,
        required=False,
    )
    entity_count = serializers.SerializerMethodField(required=False)

    def get_entity_count(self, collection: Collection) -> int:
        """Return number of entities on the collection."""
        # Use 'entity_count' attribute when available. It is created in the
        # BaseCollectionViewSet class.
        return (
            collection.entity_count
            if hasattr(collection, "entity_count")
            else collection.entity_set.count()
        )

    class Meta(BaseCollectionSerializer.Meta):
        """CollectionSerializer Meta options."""

        read_only_fields = BaseCollectionSerializer.Meta.read_only_fields + (
            "entity_count",
        )

        fields = BaseCollectionSerializer.Meta.fields + (
            "entity_count",
            "descriptor",
            "descriptor_schema",
            "descriptor_dirty",
        )



# pylint: disable=missing-docstring
from django.db import models
from versionfield import VersionField

from resolwe.flow.models.fields import ResolweSlugField


class TestModel(models.Model):
    name = models.CharField(max_length=30)

    slug = ResolweSlugField(populate_from="name", unique_with="version")

    version = VersionField(default="0.0.0")



"""Reslowe process model."""

from collections import ChainMap
from typing import Optional

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.core.validators import RegexValidator
from django.db import models

from resolwe.permissions.models import PermissionObject

from .base import BaseModel
from .data import Data


class Process(BaseModel, PermissionObject):
    """Postgres model for storing processes."""

    class Meta(BaseModel.Meta):
        """Process Meta options."""

        permissions = (
            ("view", "Can view process"),
            ("share", "Can share process"),
            ("owner", "Is owner of the process"),
        )

        indexes = [
            models.Index(name="idx_process_name", fields=["name"]),
            models.Index(name="idx_process_slug", fields=["slug"]),
            models.Index(name="idx_process_type", fields=["type"]),
            GinIndex(
                name="idx_process_type_trgm",
                fields=["type"],
                opclasses=["gin_trgm_ops"],
            ),
        ]

    #: raw persistence
    PERSISTENCE_RAW = "RAW"
    #: cached persistence
    PERSISTENCE_CACHED = "CAC"
    #: temp persistence
    PERSISTENCE_TEMP = "TMP"
    PERSISTENCE_CHOICES = (
        (PERSISTENCE_RAW, "Raw"),
        (PERSISTENCE_CACHED, "Cached"),
        (PERSISTENCE_TEMP, "Temp"),
    )

    SCHEDULING_CLASS_INTERACTIVE = "IN"
    SCHEDULING_CLASS_BATCH = "BA"
    SCHEDULING_CLASS_CHOICES = (
        (SCHEDULING_CLASS_INTERACTIVE, "Interactive"),
        (SCHEDULING_CLASS_BATCH, "Batch"),
    )

    #: data type
    type = models.CharField(
        max_length=100,
        validators=[
            RegexValidator(
                regex=r"^data:[a-z0-9:]+:$",
                message="Type may be alphanumerics separated by colon",
                code="invalid_type",
            )
        ],
    )

    #: category
    category = models.CharField(
        max_length=200,
        default="Other:",
        validators=[
            RegexValidator(
                regex=r"^([a-zA-Z0-9]+[:\-])*[a-zA-Z0-9]+:$",
                message="Category may be alphanumerics separated by colon",
                code="invalid_category",
            )
        ],
    )

    persistence = models.CharField(
        max_length=3, choices=PERSISTENCE_CHOICES, default=PERSISTENCE_RAW
    )
    """
    Persistence of :class:`~resolwe.flow.models.Data` objects created
    with this process. It can be one of the following:

    - :attr:`PERSISTENCE_RAW`
    - :attr:`PERSISTENCE_CACHED`
    - :attr:`PERSISTENCE_TEMP`

    .. note::

        If persistence is set to ``PERSISTENCE_CACHED`` or
        ``PERSISTENCE_TEMP``, the process must be idempotent.

    """

    #: designates whether this process should be treated as active
    is_active = models.BooleanField("active", default=True)

    #: detailed description
    description = models.TextField(default="")

    #: template for name of Data object created with Process
    data_name = models.CharField(max_length=200, null=True, blank=True)

    input_schema = models.JSONField(blank=True, default=list)
    """
    process input schema (describes input parameters, form layout **"Inputs"** for :attr:`Data.input`)

    Handling:

    - schema defined by: *dev*
    - default by: *user*
    - changable by: *none*

    """

    output_schema = models.JSONField(blank=True, default=list)
    """
    process output schema (describes output JSON, form layout **"Results"** for :attr:`Data.output`)

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    Implicitly defined fields (by
    :func:`resolwe.flow.management.commands.register` or
    ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its
    derivatives):

    - ``progress`` of type ``basic:float`` (from 0.0 to 1.0)
    - ``proc`` of type ``basic:group`` containing:

      - ``stdout`` of type ``basic:text``
      - ``rc`` of type ``basic:integer``
      - ``task`` of type ``basic:string`` (celery task id)
      - ``worker`` of type ``basic:string`` (celery worker hostname)
      - ``runtime`` of type ``basic:string`` (runtime instance hostname)
      - ``pid`` of type ``basic:integer`` (process ID)

    """

    entity_type = models.CharField(max_length=100, null=True, blank=True)
    """
    Automatically add :class:`~resolwe.flow.models.Data` object created
    with this process to an :class:`~resolwe.flow.models.Entity` object
    representing a data-flow. If all input ``Data`` objects belong to
    the same entity, add newly created ``Data`` object to it, otherwise
    create a new one.
    """

    entity_input = models.CharField(max_length=100, null=True, blank=True)
    """
    Limit the entity selection in
    :attr:`~resolwe.flow.models.Process.entity_type` to a single input.
    """

    entity_always_create = models.BooleanField(default=False)
    """
    Create new entity, regardless of ``entity_input`` field.
    """

    run = models.JSONField(default=dict)
    """
    process command and environment description for internal use

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    """

    requirements = models.JSONField(default=dict)
    """
    process requirements
    """

    scheduling_class = models.CharField(
        max_length=2, choices=SCHEDULING_CLASS_CHOICES, default=SCHEDULING_CLASS_BATCH
    )
    """
    process scheduling class
    """

    def get_resource_limits(self, data: Optional[Data] = None):
        """Get the core count and memory usage limits for this process.

        :return: A dictionary with the resource limits, containing the
            following keys:

            - ``memory``: Memory usage limit, in MB. Defaults to 4096 if
              not otherwise specified in the resource requirements.
            - ``cores``: Core count limit. Defaults to 1.
            - ``storage``: Size (in gibibytes) of temporary volume used for
              processing in kubernetes. Defaults to 200.

        :rtype: dict
        """
        # Known resources.
        resources = ["cores", "memory", "storage"]
        # When no resources are defined, use fallback.
        fallback = {"cores": 1, "memory": 4096, "storage": 200}

        # Prepare requirements from environment.
        environment_settings = getattr(settings, "FLOW_PROCESS_RESOURCE_OVERRIDES", {})
        environment_resources = {
            resource: environment_settings.get(resource, {}).get(self.slug)
            for resource in resources
            if environment_settings.get(resource, {}).get(self.slug)
        }

        # Gather requirements for all resources from all sources.
        # The order of requirements determines their priority.
        resources_map = ChainMap(
            data.process_resources if data is not None else {},
            environment_resources,
            self.requirements.get("resources", {}),
            getattr(settings, "FLOW_PROCESS_RESOURCE_DEFAULTS", {}),
            fallback,
        )

        # Prepare the limits and cap them.
        limits = dict()
        for resource in resources:
            limits[resource] = resources_map[resource]
            cap = getattr(settings, f"FLOW_PROCESS_MAX_{resource.upper()}", None)
            if cap is not None:
                limits[resource] = min(limits[resource], cap)

        return limits



# Generated by Django 4.2.9 on 2024-01-22 07:56

from pathlib import Path
from typing import Union

from django.db import migrations


def all_subdirectories(path: Union[str, Path]) -> set[str]:
    """Get all subdirectories of the given relative path.

    The "." is excluded.
    """
    return {f"{path}/" for path in set(Path(path).parents) - {Path(".")}}


def add_directories(apps, schema_editor):
    """Add missing directory entries to the database.

    This method is safe to run multiple times.
    """
    ReferencedPath = apps.get_model("storage", "ReferencedPath")
    StorageLocation = apps.get_model("storage", "StorageLocation")
    for storage_location in StorageLocation.all_objects.filter(status="OK").iterator():
        referenced_paths = storage_location.files.values_list("path", flat=True)
        existing_directories = {path for path in referenced_paths if path.endswith("/")}
        all_directories: set[str] = set()
        for path in referenced_paths:
            all_directories.update(all_subdirectories(path))
        if missing_directories := all_directories - existing_directories:
            created_directories = ReferencedPath.objects.bulk_create(
                ReferencedPath(path=path, size=0) for path in missing_directories
            )
            storage_location.files.add(*created_directories)


class Migration(migrations.Migration):
    dependencies = [
        ("storage", "0001_squashed_0009_referencedpath_chunk_size"),
    ]

    operations = [
        migrations.RunPython(add_directories),
    ]



"""Routing rules for websocket connections."""

from channels.routing import ChannelNameRouter, ProtocolTypeRouter, URLRouter
from django.urls import path

from .consumers import BACKGROUND_TASK_CHANNEL, BackgroundTaskConsumer, ClientConsumer

application = ProtocolTypeRouter(
    {
        # Client-facing WebSocket Consumers.
        "websocket": URLRouter(
            [path("ws/<slug:session_id>", ClientConsumer.as_asgi())]
        ),
        "channel": ChannelNameRouter(
            {
                BACKGROUND_TASK_CHANNEL: BackgroundTaskConsumer.as_asgi(),
            }
        ),
    }
)



"""
Django settings for running tests for Resolwe package.

"""

import json
import os
import sys
from distutils.util import strtobool
from importlib import metadata
from pathlib import Path

from decouple import config

PROJECT_ROOT = Path(__file__).parent.resolve()

SECRET_KEY = "secret"

DEBUG = True

MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "resolwe.auditlog.middleware.ResolweAuditMiddleware",
)

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "channels",
    # 'kombu.transport.django',  # required for Celery to work with Django DB.
    "rest_framework",
    "drf_spectacular",
    "django_filters",
    "versionfield",
    "resolwe",
    "resolwe.permissions",
    "resolwe.flow",
    "resolwe.storage",
    "resolwe.toolkit",
    "resolwe.test_helpers",
    "resolwe.observers",
    "resolwe.auditlog",
)

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
            ],
        },
    },
]

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    "resolwe.permissions.permissions.ResolwePermissionBackend",
)

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

ANONYMOUS_USER_NAME = "public"

# Get the current Tox testing environment
# NOTE: This is useful for concurrently running tests with separate environments
toxenv = os.environ.get("TOXENV", "")

# Check if PostgreSQL settings are set via environment variables
pgname = os.environ.get("RESOLWE_POSTGRESQL_NAME", "resolwe")
pguser = os.environ.get("RESOLWE_POSTGRESQL_USER", "resolwe")
pgpass = os.environ.get("RESOLWE_POSTGRESQL_PASS", "resolwe")
pghost = os.environ.get("RESOLWE_POSTGRESQL_HOST", "localhost")
pgport = int(os.environ.get("RESOLWE_POSTGRESQL_PORT", 55432))

DATABASES = {
    "default": {
        "ENGINE": "resolwe.db.postgresql",
        "NAME": pgname,
        "USER": pguser,
        "PASSWORD": pgpass,
        "HOST": pghost,
        "PORT": pgport,
        "TEST": {"NAME": pgname + "_test" + os.environ.get("PYTHONHASHSEED", "")},
    }
}

# The Redis database used by Django Channels.
REDIS_CONNECTION = {
    "host": os.environ.get("RESOLWE_REDIS_HOST", "localhost"),
    "port": int(os.environ.get("RESOLWE_REDIS_PORT", 56379)),
    "db": int(os.environ.get("RESOLWE_REDIS_DATABASE", 1)),
    "protocol": (os.environ.get("RESOLWE_REDIS_PROTOCOL", "redis")),
}
REDIS_CONNECTION_STRING = "{protocol}://{host}:{port}/{db}".format(**REDIS_CONNECTION)

LISTENER_CONNECTION = {
    # Keys in the hosts dictionary are workload connector names. Currently
    # supported are 'local', 'kubertenes', 'celery' and 'slurm'.
    "hosts": config(
        "RESOLWE_LISTENER_SERVICE_HOST",
        default='{"local": "172.17.0.1"}',
        cast=json.loads,
    ),
    "port": int(os.environ.get("RESOLWE_LISTENER_SERVICE_PORT", 53893)),
    "min_port": 50000,
    "max_port": 60000,
    "protocol": "tcp",
    # Define the max number of commands listener can process simultaneously.
    "max_concurrent_commands": config(
        "RESOLWE_LISTENER_MAX_CONCURRENT_COMMANDS", cast=int, default=10
    ),
}

# The IP address where listener is available from the communication container.
# The setting is a dictionary where key is the name of the workload connector.
COMMUNICATION_CONTAINER_LISTENER_CONNECTION = config(
    "RESOLWE_COMMUNICATION_CONTAINER_LISTENER_CONNECTION",
    default=json.dumps(LISTENER_CONNECTION["hosts"]),
    cast=json.loads,
)

#: Add affinity to the Kubernetes jobs. Example:
#: {"scheduling_class":
#:   {"batch": {"nodegroup": ["compute"]}, "interactive": {"nodegroup": ["interactive"]}}
#: The keys in the scheduling_class dictionary must correspond to the lower
#: case name of the process scheduling class.
FLOW_KUBERNETES_AFFINITY = None

# Add the settings to overcommit the CPU and memory based on the process
# scheduling type. Example:
# {"batch": {"cpu": 0.8, "memory": 0.8}}
# The default overcommitment factor is 0.8.
FLOW_KUBERNETES_OVERCOMMIT = None

FLOW_KUBERNETES_PULL_POLICY = config("RESOLWE_KUBERNETES_PULL_POLICY", default="Always")

# The  ``KUBERNETES_DISPATCHER_CONFIG_LOCATION`` specifies where the
# kubernetes workload connector reads the config from. The possible choices
# are 'incluster' and 'kubectl'. The default value is 'incluster'.
KUBERNETES_DISPATCHER_CONFIG_LOCATION = os.environ.get(
    "RESOLWE_KUBERNETES_DISPATCHER_CONFIG", "incluster"
)

# Settings in OSX/Windows are different since Docker runs in a virtual machine.
if sys.platform == "darwin":
    LISTENER_CONNECTION["hosts"]["local"] = "127.0.0.1"
    COMMUNICATION_CONTAINER_LISTENER_CONNECTION = {"local": "127.0.0.1"}

ASGI_APPLICATION = "resolwe.flow.routing.channel_routing"

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {"hosts": [REDIS_CONNECTION_STRING], "expiry": 3600},
    },
}

BROKER_URL = "django://"
CELERY_TASK_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = [CELERY_TASK_SERIALIZER]

STATIC_URL = "/static/"


# Check if any Manager settings are set via environment variables
manager_prefix = os.environ.get("RESOLWE_MANAGER_REDIS_PREFIX", "resolwe.flow.manager")
FLOW_MANAGER = {
    "REDIS_PREFIX": manager_prefix,
    "TEST": {
        "REDIS_PREFIX": manager_prefix + "-test",
    },
}

# Set custom Docker command if set via environment variable.
if "RESOLWE_DOCKER_COMMAND" in os.environ:
    FLOW_DOCKER_COMMAND = os.environ["RESOLWE_DOCKER_COMMAND"]

# Ignore errors when pulling Docker images from 'list_docker_images --pull' command.
FLOW_DOCKER_IGNORE_PULL_ERRORS = strtobool(
    os.environ.get("RESOLWE_DOCKER_IGNORE_PULL_ERRORS", "1")
)

# Don't pull Docker images if set via the environment variable.
FLOW_DOCKER_DONT_PULL = strtobool(os.environ.get("RESOLWE_DOCKER_DONT_PULL", "0"))

# Ensure all container images follow a specific format.
FLOW_CONTAINER_VALIDATE_IMAGE = r".+:(?!latest)"

FLOW_API = {
    "PERMISSIONS": "resolwe.permissions.permissions",
}
FLOW_EXPRESSION_ENGINES = [
    {
        "ENGINE": "resolwe.flow.expression_engines.jinja",
        "CUSTOM_FILTERS": [
            "resolwe.flow.tests.expression_filters",
        ],
    },
]
FLOW_EXECUTION_ENGINES = [
    "resolwe.flow.execution_engines.bash",
    "resolwe.flow.execution_engines.python",
    "resolwe.flow.execution_engines.workflow",
]

FLOW_PROCESSES_FINDERS = (
    "resolwe.flow.finders.FileSystemProcessesFinder",
    "resolwe.flow.finders.AppDirectoriesFinder",
)

FLOW_PROCESSES_RUNTIMES = ("resolwe.process.runtime.Process",)

# Which processes to schedule (if not set, all).
FLOW_PROCESSES_ALLOW_LIST = config(
    "FLOW_PROCESSES_ALLOW_LIST", default="null", cast=json.loads
)

# Which processes to ignore (by default, none).
FLOW_PROCESSES_IGNORE_LIST = config(
    "FLOW_PROCESSES_IGNORE_LIST", default="null", cast=json.loads
)

FLOW_DOCKER_NETWORK = config("RESOLWE_DOCKER_NETWORK", default="bridge")

FLOW_EXECUTOR = {
    "NAME": "resolwe.flow.executors.docker",
    "LISTENER_CONNECTION": LISTENER_CONNECTION,
    "NETWORK": FLOW_DOCKER_NETWORK,
}

FLOW_DOCKER_AUTOREMOVE = False
FLOW_DOCKER_COMMUNICATOR_IMAGE = os.environ.get(
    "RESOLWE_COMMUNICATOR_IMAGE", "public.ecr.aws/s4q6j6e8/resolwe/com:latest"
)
FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE = (
    "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04"
)

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
    ),
    "DEFAULT_FILTER_BACKENDS": (
        "resolwe.permissions.filters.ResolwePermissionsFilter",
        "django_filters.rest_framework.DjangoFilterBackend",
        "resolwe.flow.filters.OrderingFilter",
    ),
    "EXCEPTION_HANDLER": "resolwe.flow.utils.exceptions.resolwe_exception_handler",
    "DATETIME_FORMAT": "iso-8601",
    "DATE_FORMAT": "iso-8601",
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
}


# Schema generation

SPECTACULAR_SETTINGS = {
    "TITLE": "Resolwe",
    "DESCRIPTION": "An open source dataflow package for Django.",
    "VERSION": metadata.metadata("resolwe")["version"],
    "SERVE_INCLUDE_SCHEMA": False,
    "ENUM_NAME_OVERRIDES": {"DataStatus": "resolwe.flow.models.Data.STATUS_CHOICES"},
}


# Time
USE_TZ = True
TIME_ZONE = "UTC"

# Django does not support parsing of 'iso-8601' formated datetimes by default.
# Since Django-filters uses Django forms for parsing, we need to modify Django
# setting ``DATETIME_INPUT_FORMATS`` to support 'iso-8601' format.
# https://docs.djangoproject.com/en/1.11/ref/settings/#datetime-input-formats
DATETIME_INPUT_FORMATS = (
    # These are already given Django defaults:
    "%Y-%m-%d %H:%M:%S",  # '2006-10-25 14:30:59'
    "%Y-%m-%d %H:%M:%S.%f",  # '2006-10-25 14:30:59.000200'
    "%Y-%m-%d %H:%M",  # '2006-10-25 14:30'
    "%Y-%m-%d",  # '2006-10-25'
    # These are iso-8601 formatted:
    "%Y-%m-%dT%H:%M:%S.%f%z",  # '2006-10-25T14:30:59.000200+0200' or '2006-10-25T14:30:59.000200+02:00' (Python>=3.7)
    "%Y-%m-%dT%H:%M:%S.%fZ",  # '2006-10-25T14:30:59.000200Z'
    "%Y-%m-%dT%H:%M:%S.%f",  # '2006-10-25T14:30:59.000200'
    "%Y-%m-%dT%H:%M:%SZ",  # '2006-10-25T14:30:59Z'
    "%Y-%m-%dT%H:%M:%S",  # '2006-10-25T14:30:59'
    "%Y-%m-%dT%H:%M",  # '2006-10-25T14:30'
)


# Testing.

TEST_RUNNER = "resolwe.test_helpers.test_runner.ResolweRunner"
TEST_PROCESS_REQUIRE_TAGS = True
TEST_PROCESS_PROFILE = False


# Logging.

# Set RESOLWE_LOG_FILE environment variable to a file path to enable
# logging debugging messages to to a file.
debug_file_path = os.environ.get("RESOLWE_LOG_FILE", os.devnull)

github_actions_debug = strtobool(os.environ.get("GITHUB_ACTIONS_DEBUG", "0"))
CONSOLE_LEVEL = "WARNING"
default_logger_handlers = ["file"]

if github_actions_debug:
    CONSOLE_LEVEL = "DEBUG"
    default_logger_handlers = ["console", "file"]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s - %(levelname)s - %(name)s[%(process)s]: %(message)s",
        },
        "auditlog": {
            "format": "%(asctime)s - %(name)s[%(process)s]: %(message)s ; UID='%(user_id)s' SID='%(session_id)s' RID='%(request_id)s'",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": CONSOLE_LEVEL,
            "formatter": "standard",
        },
        "auditlog": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "auditlog",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": debug_file_path,
            "formatter": "standard",
            "maxBytes": 1024 * 1024 * 10,  # 10 MB
        },
    },
    "loggers": {
        "": {
            "handlers": default_logger_handlers,
            "level": "DEBUG",
        },
        "auditlog": {
            "handlers": ["auditlog"],
            "level": "INFO",
        },
    },
}



"""Listener authenticator."""

import logging
from collections import defaultdict

from channels.db import database_sync_to_async
from zmq.auth.asyncio import AsyncioAuthenticator

from resolwe.flow.models import Worker
from resolwe.flow.utils import Singleton

logger = logging.getLogger(__name__)


class ZMQAuthenticator(AsyncioAuthenticator, Singleton):
    """The singleton AsyncioAuthenticator authenticator.

    It is also used for handling curve callbacks.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the mapping between key and the data objects.

        The key is the public key of the worker, and the value is the list of data ids
        the worker is allowed to access.
        """
        super().__init__(*args, **kwargs)
        self._authorizations: dict[bytes, set[int]] = defaultdict(set)
        self.configure_curve_callback(domain="*", credentials_provider=self)

    async def callback(self, domain: str, worker_public_key: bytes) -> bool:
        """Authenticate the client.

        Every client is accepted. On first connection, its key is added to the
        authorization dictionary.
        """

        def get_data_ids(worker_public_key: bytes) -> set[int]:
            """Get the data ids the worker is allowed to access."""
            acceptable_worker_statuses = [
                Worker.STATUS_PREPARING,
                Worker.STATUS_PROCESSING,
                Worker.STATUS_FINISHED_PREPARING,
            ]
            results = Worker.objects.filter(public_key=worker_public_key).values_list(
                "status", "data_id"
            )
            return {
                data_id
                for status, data_id in results
                if status in acceptable_worker_statuses
            }

        try:
            assert domain == "*", "Only domain '*' is supported."

            data_ids = await database_sync_to_async(
                get_data_ids, thread_sensitive=False
            )(worker_public_key)
            self._authorizations[worker_public_key] = data_ids

        except Exception:
            logger.exception("Exception in curve auth callback.")

        return True

    def can_access_data(self, worker_public_key: bytes, data_id: int) -> bool:
        """Can worker access the given data object."""
        return data_id in self._authorizations.get(worker_public_key, [])

    def clear_authorizations(self):
        """Clear the authorization dictionary."""
        self._authorizations.clear()

    def authorize_client(self, worker_public_key: bytes, data_id: int):
        """Authorize the client to access the data object."""
        self._authorizations[worker_public_key].add(data_id)



# pylint: disable=missing-docstring
import hashlib
import os
import shutil
import tempfile

import crcmod
from django.conf import settings

from resolwe.storage.connectors import (
    AwsS3Connector,
    GoogleConnector,
    LocalFilesystemConnector,
)
from resolwe.storage.connectors.transfer import Transfer
from resolwe.storage.connectors.utils import get_transfer_object
from resolwe.test import TestCase

CONNECTORS = {
    "local": {
        "path": "CHANGE",
    },
    "S3": {
        "bucket": "genialis-test-storage",
        "credentials": os.path.join(
            settings.PROJECT_ROOT, "testing_credentials_s3.json"
        ),
    },
    "GCS": {
        "bucket": "genialis_storage_test",
        "credentials": os.path.join(
            settings.PROJECT_ROOT, "testing_credentials_gcs.json"
        ),
    },
}


def hashes(filename):
    """Compute MD5 and CRC32C hexdigest."""
    hashers = (hashlib.md5(), crcmod.predefined.PredefinedCrc("crc32c"))
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            for hasher in hashers:
                hasher.update(chunk)
    return {"md5": hashers[0].hexdigest(), "crc32c": hashers[1].hexdigest()}


def create_file(name, size):
    with open(name, "wb") as f:
        f.seek(size - 1)
        f.write(b"\0")


class BaseTestCase(TestCase):
    _mb = 1024 * 1024
    _files = (
        ("1", 1 * _mb),
        ("dir/2", 10 * _mb),
    )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.tmp_dir = tempfile.TemporaryDirectory()
        local_settings = CONNECTORS["local"].copy()
        local_settings["path"] = cls.tmp_dir.name
        cls.created_files = [f[0] for f in cls._files]
        cls.local = LocalFilesystemConnector(local_settings, "local")
        cls.gcs = GoogleConnector(CONNECTORS["GCS"], "GCS")
        cls.s3 = AwsS3Connector(CONNECTORS["S3"], "S3")

    @classmethod
    def _purge_tmp_dir(cls):
        for root, dirs, files in os.walk(cls.tmp_dir.name):
            for f in files:
                os.unlink(os.path.join(root, f))
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))

    @classmethod
    def _create_files(cls):
        os.makedirs(cls.path(cls, ""))
        os.makedirs(cls.path(cls, "dir"))
        cls.transfer_objects = []
        for file_, size in cls._files:
            create_file(cls.path(cls, file_), size)
            with open(cls.path(cls, file_), "r+b") as f:
                f.write((file_.encode("utf-8")))
            cls.transfer_objects.append(
                get_transfer_object(
                    cls.path(cls, file_), os.path.join(cls.tmp_dir.name, cls._prefix)
                )
            )

    @classmethod
    def _copy_files_to(cls, connector):
        for file_ in cls.created_files:
            connector_path = os.path.join(cls._prefix, file_)
            if not connector.exists(connector_path):
                with open(cls.path(cls, file_), "rb") as f:
                    connector.push(f, connector_path)
                    connector.set_hashes(connector_path, hashes(cls.path(cls, file_)))

    def path(self, filename):
        return os.path.join(self.tmp_dir.name, self._prefix, filename)

    def _test_from_to(self, from_connector, to_connector):
        t = Transfer(from_connector, to_connector)
        t.transfer_objects(self._prefix, self.transfer_objects)
        objects = to_connector.get_object_list(self._prefix)
        self.assertEqual(len(objects), len(self.created_files))
        for file_ in self.created_files:
            real_path = os.path.join(self._prefix, file_)
            self.assertIn(file_, objects)
            for type_ in ("md5", "crc32c"):
                self.assertEqual(
                    from_connector.get_hash(real_path, type_).lower(),
                    to_connector.get_hash(real_path, type_).lower(),
                )


class TransferFromLocalTest(BaseTestCase):
    _prefix = "local"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._create_files()
        cls.gcs.delete(cls._prefix, cls.created_files)
        cls.s3.delete(cls._prefix, cls.created_files)

    def test_to_google(self):
        self._test_from_to(self.local, self.gcs)

    def test_to_s3(self):
        self._test_from_to(self.local, self.s3)


class TransferFromGoogleTest(BaseTestCase):
    _prefix = "google"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._create_files()
        cls.gcs.delete(cls._prefix, cls.created_files)
        cls._copy_files_to(cls.gcs)
        cls._purge_tmp_dir()

    def setUp(self):
        super().setUp()
        self.local.delete(self._prefix, self.created_files)
        self.s3.delete(self._prefix, self.created_files)

    def test_to_local(self):
        self._test_from_to(self.gcs, self.local)

    def test_to_s3(self):
        self._test_from_to(self.gcs, self.s3)


class TransferFromS3Test(BaseTestCase):
    _prefix = "s3"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._create_files()
        cls.s3.delete(cls._prefix, cls.created_files)
        cls._copy_files_to(cls.s3)
        cls._purge_tmp_dir()

    def setUp(self):
        super().setUp()
        self.gcs.delete(self._prefix, self.created_files)
        self.local.delete(self._prefix, self.created_files)

    def test_to_local(self):
        self._test_from_to(self.s3, self.local)

    def test_to_gcs(self):
        self._test_from_to(self.s3, self.gcs)



"""Data viewset."""

from django.db.models import Prefetch
from drf_spectacular.utils import extend_schema
from rest_framework import (
    exceptions,
    mixins,
    permissions,
    serializers,
    status,
    viewsets,
)
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.filters import DataFilter
from resolwe.flow.models import Data, DescriptorSchema, Process
from resolwe.flow.models.utils import fill_with_defaults
from resolwe.flow.serializers import DataSerializer
from resolwe.flow.utils import get_data_checksum
from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.views import BackgroundTaskSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import Permission, PermissionModel

from .collection import BaseCollectionViewSet
from .entity import EntityViewSet
from .mixins import (
    ResolweBackgroundDeleteMixin,
    ResolweBackgroundDuplicateMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)
from .utils import get_collection_for_user


class IsStaffuser(permissions.BasePermission):
    """Allow access only to staff users."""

    message = "Only staff users are allowed."

    def has_permission(self, request, view):
        """Return true when request is allowed."""
        return bool(
            request.user and request.user.is_authenticated and request.user.is_staff
        )


class MoveDataToCollectionSerializer(serializers.Serializer):
    """Deserializer for data move to collection endpoint."""

    ids = serializers.ListField(child=serializers.IntegerField())
    destination_collection = serializers.IntegerField()


class RestartSerializer(serializers.Serializer):
    """Serializer for restarting a Data object."""

    resource_overrides = serializers.JSONField(required=False, default=dict)


class DataViewSet(
    ObservableMixin,
    ResolweCreateModelMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    ResolweBackgroundDeleteMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ResolweBackgroundDuplicateMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`Data` objects."""

    qs_permission_model = PermissionModel.objects.select_related("user", "group")
    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")
    qs_process = Process.objects.select_related("contributor")
    queryset = Data.objects.select_related("contributor").prefetch_related(
        Prefetch("collection", queryset=BaseCollectionViewSet.queryset),
        Prefetch("descriptor_schema", queryset=qs_descriptor_schema),
        Prefetch("entity", queryset=EntityViewSet.queryset),
        Prefetch("process", queryset=qs_process),
    )

    serializer_class = DataSerializer
    filterset_class = DataFilter
    permission_classes = (get_permissions_class(),)

    ordering_fields = (
        "contributor",
        "contributor__first_name",
        "contributor__last_name",
        "created",
        "finished",
        "id",
        "modified",
        "name",
        "process__name",
        "process__type",
        "started",
        "entity__name",
    )
    ordering = "-created"

    def get_queryset(self):
        """Prefetch permissions for current user."""
        return self.prefetch_current_user_permissions(self.queryset)

    @action(detail=False, methods=["post"])
    def get_or_create(self, request, *args, **kwargs):
        """Get ``Data`` object if similar already exists, otherwise create it."""
        response = self.perform_get_or_create(request, *args, **kwargs)
        if response:
            return response

        return super().create(request, *args, **kwargs)

    def perform_get_or_create(self, request, *args, **kwargs):
        """Perform "get_or_create" - return existing object if found."""
        self.define_contributor(request)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        process = serializer.validated_data.get("process")
        process_input = request.data.get("input", {})

        fill_with_defaults(process_input, process.input_schema)

        checksum = get_data_checksum(process_input, process.slug, process.version)
        data_qs = Data.objects.filter(
            checksum=checksum,
            process__persistence__in=[
                Process.PERSISTENCE_CACHED,
                Process.PERSISTENCE_TEMP,
            ],
        )
        data_qs = data_qs.filter_for_user(request.user)
        if data_qs.exists():
            data = data_qs.order_by("created").last()
            serializer = self.get_serializer(data)
            return Response(serializer.data)

    def _parents_children(self, request, queryset):
        """Process given queryset and return serialized objects."""
        queryset = queryset.filter_for_user(request.user)
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    @extend_schema(
        filters=False, responses={status.HTTP_200_OK: DataSerializer(many=True)}
    )
    @action(detail=True)
    def parents(self, request, pk=None):
        """Return parents of the current data object."""
        return self._parents_children(request, self.get_object().parents)

    @extend_schema(
        filters=False,
        request=RestartSerializer(),
        responses={status.HTTP_200_OK: DataSerializer(many=True)},
    )
    @action(detail=True, methods=["post"], permission_classes=[IsStaffuser])
    def restart(self, request, pk=None):
        """Restart the current data object."""
        argument_validator = RestartSerializer(data=request.data)
        argument_validator.is_valid(raise_exception=True)
        data = self.get_object()
        try:
            data.restart(**argument_validator.validated_data)
        except RuntimeError as e:
            raise exceptions.ValidationError(str(e))
        return Response(self.get_serializer(data).data)

    @action(detail=True)
    def children(self, request, pk=None):
        """Return children of the current data object."""
        return self._parents_children(request, self.get_object().children.all())

    @extend_schema(
        request=MoveDataToCollectionSerializer(),
        responses={status.HTTP_200_OK: None},
    )
    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move data objects to destination collection."""
        serializer = MoveDataToCollectionSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        ids = serializer.validated_data["ids"]
        dst_collection_id = serializer.validated_data["destination_collection"]

        dst_collection = get_collection_for_user(dst_collection_id, request.user)

        queryset = self._get_data(request.user, ids)
        task = queryset.move_to_collection(dst_collection, request.user)
        return Response(
            status=status.HTTP_200_OK,
            data=BackgroundTaskSerializer(task).data,
        )

    def _get_data(self, user, ids):
        """Return data objects queryset based on provided ids."""
        queryset = Data.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Data objects with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        for data in queryset:
            collection = data.collection
            if collection and not user.has_perm(Permission.EDIT, obj=collection):
                if user.is_authenticated:
                    raise exceptions.PermissionDenied()
                else:
                    raise exceptions.NotFound()

        return queryset



""".. Ignore pydocstyle D400.

===============
Signal Handlers
===============

"""

import logging

from asgiref.sync import async_to_sync
from channels.layers import ChannelFull, get_channel_layer
from django.db.models.signals import post_delete
from django.dispatch import receiver

from resolwe.flow.models import Data
from resolwe.storage.protocol import (
    CHANNEL_STORAGE_CLEANUP_WORKER,
    TYPE_STORAGE_CLEANUP_RUN,
)

logger = logging.getLogger(__name__)


@receiver(post_delete, sender=Data)
def data_post_delete_handler(sender, instance: Data, using, **kwargs):
    """Remove files referenced by deleted object."""
    channel_layer = get_channel_layer()
    channel_data = {"type": TYPE_STORAGE_CLEANUP_RUN}
    try:
        if instance.location_id is not None:
            channel_data["file_storage_id"] = instance.location_id
        async_to_sync(channel_layer.send)(CHANNEL_STORAGE_CLEANUP_WORKER, channel_data)
    except ChannelFull:
        logger.warning("Cannot trigger storage manager run because channel is full.")



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.test.ProcessTestCase
    :members:

"""

import contextlib
import filecmp
import gzip
import hashlib
import io
import json
import os
import shutil
import sys
import tarfile
import tempfile
import time
import uuid
import zipfile
from itertools import filterfalse

from django.conf import settings
from django.core import management
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import transaction
from django.utils.text import slugify

from resolwe.flow.models import Data, DescriptorSchema, Process, Storage
from resolwe.flow.utils import dict_dot, iterate_fields, iterate_schema
from resolwe.permissions.models import PermissionGroup
from resolwe.storage.connectors import connectors
from resolwe.test import TransactionTestCase

from ..utils import get_processes_from_tags, has_process_tag

SCHEMAS_FIXTURE_CACHE = None


class TestProfiler:
    """Simple test profiler."""

    def __init__(self, test):
        """Initialize test profiler.

        :param test: Unit test instance
        """
        self._test = test
        self._start = time.time()

        if getattr(settings, "TEST_PROCESS_PROFILE", False):
            self._file = open(
                "profile-resolwe-process-tests-{}.json".format(os.getpid()), "a"
            )
        else:
            self._file = None

        # Automatically cleanup when test completes.
        test.addCleanup(self.close)

    def add(self, data):
        """Add output to profile log.

        :param data: Arbitrary data dictionary
        """
        if not self._file:
            return

        data.update({"test": self._test.id()})

        self._file.write(json.dumps(data))
        self._file.write("\n")

    @contextlib.contextmanager
    def add_block(self, name):
        """Profile a named block of code.

        :param name: Block name
        """
        block_start = time.time()
        try:
            yield
        finally:
            block_end = time.time()
            self.add({name: block_end - block_start})

    def close(self):
        """Close profiler log."""
        if not self._file:
            return

        self.add({"total": time.time() - self._start})
        self._file.close()


class ProcessTestCase(TransactionTestCase):
    """Base class for writing process tests.

    It is a subclass of :class:`.TransactionTestCase` with some specific
    functions used for testing processes.

    To write a process test use standard Django's syntax for writing
    tests and follow the next steps:

    #. Put input files (if any) in ``tests/files`` directory of a
       Django application.
    #. Run the process using
       :meth:`.run_process`.
    #. Check if the process has the expected status using
       :meth:`.assertStatus`.
    #. Check process's output using :meth:`.assertFields`,
       :meth:`.assertFile`, :meth:`.assertFileExists`,
       :meth:`.assertFiles` and :meth:`.assertJSON`.

    .. note::
        When creating a test case for a custom Django application,
        subclass this class and over-ride the ``self.files_path`` with:

        .. code-block:: python

            self.files_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')

    .. DANGER::
        If output files don't exist in ``tests/files`` directory of a
        Django application, they are created automatically.
        But you have to check that they are correct before using them
        for further runs.

    """

    def _update_schema_relations(self, schemas):
        """Update foreign keys on process and descriptor schema.

        The field contributor is updated.

        """
        for schema in schemas:
            schema.contributor = self.admin

    def _create_permission_groups(self, schemas):
        """Create permission groups."""
        for schema in schemas:
            schema.permission_group = PermissionGroup.objects.create()

    def _register_schemas(
        self, path=None, processes_paths=None, descriptors_paths=None
    ):
        """Register process and descriptor schemas.

        If ``processes_paths`` or ``descriptor_path`` arguments are
        given, only processes/descriptors from that path are registered.

        ``path`` argument is supported for backward compatibility and
        applies to both arguments above.

        Process and DescriptorSchema are cached to SCHEMAS_FIXTURE_CACHE
        global variable based on ``processes_paths`` and
        ``descriptors_paths`` keys in ``kwargs``.

        """

        def remove_pks(schemas):
            """Remove primary keys from the given schemas."""
            for s in schemas:
                s.pk = None
            return schemas

        if path:
            if processes_paths or descriptors_paths:
                raise ValueError(
                    "processes_paths and descriptors_paths arguments must be None if path is defined."
                )
            processes_paths = descriptors_paths = path

        schemas_types = [
            {
                "name": "descriptor_schemas",
                "model": DescriptorSchema,
                "cache_key": descriptors_paths,
            },
            {
                "name": "processes",
                "model": Process,
                "cache_key": processes_paths,
            },
        ]

        for schemas in schemas_types:
            schemas["model"].objects.all().delete()

        cache_key = str(
            "processes_paths={};descriptors_paths={}".format(
                processes_paths, descriptors_paths
            )
        )
        global SCHEMAS_FIXTURE_CACHE
        if not SCHEMAS_FIXTURE_CACHE:
            SCHEMAS_FIXTURE_CACHE = {}

        stdout, stderr = io.StringIO(), io.StringIO()

        if cache_key in SCHEMAS_FIXTURE_CACHE:
            for schemas in schemas_types:
                # NOTE: Schemas' current primary keys may not be unique on the next runs of
                # processes' tests, therefore we must remove them and let the DB re-create them
                # properly.
                # WARNING: Cached schemas' primary keys will be set on every call to bulk_create(),
                # therefore we need to remove them and let the DB re-create them every time. For
                # more details, see:
                # https://github.com/django/django/blob/1.10.7/django/db/models/query.py#L455-L456
                schemas_cache = remove_pks(
                    SCHEMAS_FIXTURE_CACHE[cache_key][schemas["name"]]
                )
                self._update_schema_relations(schemas_cache)
                self._create_permission_groups(schemas_cache)
                schemas["model"].objects.bulk_create(schemas_cache)
        else:
            if processes_paths is None and descriptors_paths is None:
                management.call_command(
                    "register", force=True, stdout=stdout, stderr=stderr
                )
            else:
                settings_overrides = {
                    "FLOW_PROCESSES_FINDERS": [
                        "resolwe.flow.finders.FileSystemProcessesFinder"
                    ]
                }
                if processes_paths is not None:
                    settings_overrides["FLOW_PROCESSES_DIRS"] = processes_paths
                if descriptors_paths is not None:
                    settings_overrides["FLOW_DESCRIPTORS_DIRS"] = descriptors_paths

                with self.settings(**settings_overrides):
                    management.call_command(
                        "register", force=True, stdout=stdout, stderr=stderr
                    )

            if cache_key not in SCHEMAS_FIXTURE_CACHE:
                SCHEMAS_FIXTURE_CACHE[cache_key] = {}

            # NOTE: list() forces DB query execution
            for schemas in schemas_types:
                SCHEMAS_FIXTURE_CACHE[cache_key][schemas["name"]] = list(
                    schemas["model"].objects.all()
                )

        return stdout, stderr

    def setUp(self):
        """Initialize test data."""
        super().setUp()

        _, stderr = self._register_schemas()
        stderr = stderr.getvalue()
        if stderr:
            raise RuntimeError(stderr)

        upload_connectors = [
            connector
            for connector in connectors.for_storage("upload")
            if connector.mountable
        ]
        assert upload_connectors, "No upload connector defined on filesystem"

        self.upload_dir = upload_connectors[0].path
        self._profiler = TestProfiler(self)
        self._preparation_stage = 0
        self._executed_processes = set()
        self._files_path = None
        self._upload_files = []

        # create upload dir if it doesn't exist
        if not os.path.isdir(self.upload_dir):
            os.mkdir(self.upload_dir)

    def tearDown(self):
        """Clean up after the test."""
        # delete Data objects and their files unless keep_data
        for d in Data.objects.all():
            if self._keep_data:
                print("KEEPING DATA: {}".format(d.pk))
            elif d.location:
                data_dir = d.location.get_path()
                shutil.rmtree(data_dir, ignore_errors=True)
                d.delete()

        # remove uploaded files
        if not self._keep_data:
            for fn in self._upload_files:
                shutil.rmtree(fn, ignore_errors=True)

        super().tearDown()

        # Check test outcome to prevent failing the test twice.
        # Adapted from: https://stackoverflow.com/a/39606065
        def list2reason(exc_list):
            """Error reason conversion helper."""
            if exc_list and exc_list[-1][0] is self:
                return exc_list[-1][1]

        result = self.defaultTestResult()

        # Python 3.11 has no feedErrorsToResult method.
        if sys.version_info.major == 3 and sys.version_info.minor <= 10:
            self._feedErrorsToResult(result, self._outcome.errors)
        error = list2reason(result.errors)
        failure = list2reason(result.failures)

        # Ensure all tagged processes were tested.
        if (
            not error
            and not failure
            and getattr(settings, "TEST_PROCESS_REQUIRE_TAGS", False)
        ):
            test = getattr(self, self._testMethodName)
            for slug in get_processes_from_tags(test):
                if slug not in self._executed_processes:
                    self.fail(
                        'Test was tagged with process "{}", but this process was not '
                        "executed during test. Remove the tag or test the process.".format(
                            slug
                        )
                    )

    @contextlib.contextmanager
    def preparation_stage(self):
        """Context manager to mark input preparation stage."""
        with self._profiler.add_block("preparation"):
            self._preparation_stage += 1
            try:
                yield
            finally:
                self._preparation_stage -= 1

        # TODO: Handle automatic caching.

    @property
    def files_path(self):
        """Path to test files."""
        if self._files_path is None:
            raise NotImplementedError

        return self._files_path

    @files_path.setter
    def files_path(self, value):
        self._files_path = value

    def run_processor(self, *args, **kwargs):
        """Run process.

        Deprecated method: use run_process.

        """
        return self.run_process(*args, **kwargs)
        # TODO: warning

    def run_process(
        self,
        process_slug,
        input_={},
        assert_status=Data.STATUS_DONE,
        descriptor=None,
        descriptor_schema=None,
        verbosity=0,
        tags=None,
        contributor=None,
        collection=None,
        process_resources=None,
    ):
        """Run the specified process with the given inputs.

        If input is a file, file path should be given relative to the
        ``tests/files`` directory of a Django application.
        If ``assert_status`` is given, check if
        :class:`~resolwe.flow.models.Data` object's status matches
        it after the process has finished.

        .. note::

            If you need to delay calling the manager, you must put the
            desired code in a ``with transaction.atomic()`` block.

        :param str process_slug: slug of the
            :class:`~resolwe.flow.models.Process` to run

        :param dict ``input_``: :class:`~resolwe.flow.models.Process`'s
            input parameters

            .. note::

                You don't have to specify parameters with defined
                default values.

        :param str ``assert_status``: desired status of the
            :class:`~resolwe.flow.models.Data` object

        :param dict descriptor: descriptor to set on the
            :class:`~resolwe.flow.models.Data` object

        :param dict descriptor_schema: descriptor schema to set on the
            :class:`~resolwe.flow.models.Data` object

        :param list tags: list of tags that will be added to the created
            :class:`~resolwe.flow.models.Data` object

        :return: object created by
            :class:`~resolwe.flow.models.Process`
        :rtype: ~resolwe.flow.models.Data

        """
        # Copy input_, to avoid mutation that would occur in ``mock_upload``
        input_ = input_.copy()

        # backward compatibility
        process_slug = slugify(process_slug.replace(":", "-"))

        # Enforce correct process tags.
        if (
            getattr(settings, "TEST_PROCESS_REQUIRE_TAGS", False)
            and not self._preparation_stage
        ):
            test = getattr(self, self._testMethodName)
            if not has_process_tag(test, process_slug):
                self.fail(
                    'Tried to run process with slug "{0}" outside of preparation_stage\n'
                    "block while test is not tagged for this process. Either tag the\n"
                    "test using tag_process decorator or move this under the preparation\n"
                    "stage block if this process is only used to prepare upstream inputs.\n"
                    "\n"
                    "To tag the test you can add the following decorator:\n"
                    "    @tag_process('{0}')\n"
                    "".format(process_slug)
                )

        self._executed_processes.add(process_slug)

        process = Process.objects.filter(slug=process_slug).order_by("-version").first()

        if process is None:
            self.fail('No process with slug "{}"'.format(process_slug))

        def mock_upload(file_path):
            """Mock file upload."""

            def is_url(path):
                """Check if path is a URL."""
                validate = URLValidator()
                try:
                    validate(path)
                except (ValueError, ValidationError):
                    return False
                return True

            if is_url(file_path):
                return {
                    "file": file_path,
                    "file_temp": file_path,
                    "is_remote": True,
                }
            else:
                old_path = os.path.join(self.files_path, file_path)
                if not os.path.isfile(old_path):
                    raise RuntimeError("Missing file: {}".format(old_path))

                file_basename = os.path.basename(file_path)

                file_temp = "{}_{}".format(file_basename, uuid.uuid4())
                upload_file_path = os.path.join(self.upload_dir, file_temp)
                # create directories needed by new_path
                upload_file_dir = os.path.dirname(upload_file_path)
                if not os.path.exists(upload_file_dir):
                    os.makedirs(upload_file_dir)

                shutil.copy2(old_path, upload_file_path)
                self._upload_files.append(upload_file_path)
                return {
                    "file": file_basename,
                    "file_temp": file_temp,
                }

        for field_schema, fields in iterate_fields(input_, process.input_schema):
            # copy referenced files to upload dir
            if field_schema["type"] == "basic:file:":
                fields[field_schema["name"]] = mock_upload(fields[field_schema["name"]])
            elif field_schema["type"] == "list:basic:file:":
                file_list = [
                    mock_upload(file_path) for file_path in fields[field_schema["name"]]
                ]
                fields[field_schema["name"]] = file_list

            # convert primary keys to strings
            if field_schema["type"].startswith("data:"):
                fields[field_schema["name"]] = fields[field_schema["name"]]
            if field_schema["type"].startswith("list:data:"):
                fields[field_schema["name"]] = [
                    obj for obj in fields[field_schema["name"]]
                ]

        data = Data.objects.create(
            input=input_,
            contributor=contributor or self.admin,
            process=process,
            tags=tags or [],
            descriptor_schema=descriptor_schema,
            descriptor=descriptor or {},
            collection=collection,
            process_resources=process_resources or {},
        )

        # Fetch latest Data object from database
        data = Data.objects.get(pk=data.pk)

        if assert_status:
            if not transaction.get_autocommit() and assert_status == Data.STATUS_DONE:
                # We are in an atomic transaction block, hence the data object will not be done
                # until after the block. Therefore the expected status is resolving.
                assert_status = Data.STATUS_RESOLVING
            self.assertStatus(data, assert_status)

        return data

    def get_json(self, file_name, storage):
        """Return JSON saved in file and test JSON to compare it to.

        The method returns a tuple of the saved JSON and the test JSON.
        In your test you should then compare the test JSON to the saved
        JSON that is commited to the repository.

        The storage argument could be a Storage object, Storage ID or a
        Python dictionary. The test JSON is assigned a json field of
        the Storage object or the complete Python dictionary
        (if a dict is given).

        If the file does not exist it is created, the test JSON is
        written to the new file and an exception is rased.

        :param str file_name: file name (and relative path) of a JSON
            file. Path should be relative to the ``tests/files``
            directory of a Django app. The file name must have a ``.gz`` extension.
        :param storage: Storage object, Storage ID or a dict.
        :type storage: :class:`~resolwe.flow.models.Storage`,
            :class:`str` or :class:`dict`
        :return: (reference JSON, test JSON)
        :rtype: tuple

        """
        self.assertEqual(
            os.path.splitext(file_name)[1], ".gz", msg="File extension must be .gz"
        )

        if isinstance(storage, Storage):
            json_dict = storage.json
        elif isinstance(storage, int):
            json_dict = Storage.objects.get(pk=storage).json
        elif isinstance(storage, dict):
            json_dict = storage
        else:
            raise ValueError("Argument storage should be of type Storage, int or dict.")

        file_path = os.path.join(self.files_path, file_name)
        if not os.path.isfile(file_path):
            with gzip.open(file_path, mode="wt") as f:
                json.dump(json_dict, f)

            self.fail(msg="Output file {} missing so it was created.".format(file_name))

        with gzip.open(file_path, mode="rt") as f:
            return json.load(f), json_dict

    def assertStatus(self, obj, status):
        """Check if object's status is equal to the given status.

        :param obj: object for which to check the status
        :type obj: ~resolwe.flow.models.Data
        :param str status: desired value of object's
            :attr:`~resolwe.flow.models.Data.status` attribute

        """
        self.assertEqual(
            obj.status,
            status,
            msg="Data status is '{}', not '{}'".format(obj.status, status)
            + self._debug_info(obj),
        )

    def _get_output_field(self, obj, path):
        """Return object's output field schema and field dict.

        :param obj: object with the output field
        :type obj: ~resolwe.flow.models.Data

        :param str path: path to :class:`~resolwe.flow.models.Data`
            object's output field

        """
        for field_schema, field, field_path in iterate_fields(
            obj.output, obj.process.output_schema, ""
        ):
            if path == field_path:
                return field_schema, field

        self.fail("Field not found in path {}.".format(path))

    def assertFields(self, obj, path, value):
        """Compare object's field to the given value.

        The file size is ignored. Use assertFile to validate
        file contents.

        :param obj: object with the field to compare
        :type obj: ~resolwe.flow.models.Data

        :param str path: path to
            :class:`~resolwe.flow.models.Data` object's field

        :param str value: desired value of
            :class:`~resolwe.flow.models.Data` object's field

        """
        field_schema, field = None, None
        for field_schema, field, field_path in iterate_schema(
            obj.output, obj.process.output_schema, ""
        ):
            if path == field_path:
                break
        else:
            self.fail("Field not found in path {}".format(path))

        field_name = field_schema["name"]
        field_value = field[field_name]

        def remove_file_size(field_value):
            """Remove size value from file field."""
            if "size" in field_value:
                del field_value["size"]

        # Ignore size in file and dir fields
        if field_schema["type"].startswith("basic:file:") or field_schema[
            "type"
        ].startswith("basic:dir:"):
            remove_file_size(field_value)
            remove_file_size(value)

        elif field_schema["type"].startswith("list:basic:file:") or field_schema[
            "type"
        ].startswith("list:basic:dir:"):
            for val in field_value:
                remove_file_size(val)
            for val in value:
                remove_file_size(val)

        self.assertEqual(
            field_value,
            value,
            msg="Field 'output.{}' mismatch: {} != {}".format(path, field_value, value)
            + self._debug_info(obj),
        )

    def _assert_file(
        self,
        obj,
        fn_tested,
        fn_correct,
        compression=None,
        file_filter=lambda _: False,
        sort=False,
    ):
        """Compare files."""
        open_kwargs = {}
        if compression is None:
            open_fn = open
            # by default, open() will open files as text and return str
            # objects, but we need bytes objects
            open_kwargs["mode"] = "rb"
        elif compression == "gzip":
            open_fn = gzip.open
        elif compression == "zip":
            open_fn = zipfile.ZipFile.open
        else:
            raise ValueError("Unsupported compression format.")

        def get_sha256(filename, **kwargs):
            """Get sha256 for a given file."""
            with open_fn(filename, **kwargs) as handle:
                contents = [line for line in filterfalse(file_filter, handle)]
                if sort:
                    contents = sorted(contents)
                contents = b"".join(contents)
            return hashlib.sha256(contents).hexdigest()

        output = obj.location.get_path(filename=fn_tested)
        output_hash = get_sha256(output, **open_kwargs)

        correct_path = os.path.join(self.files_path, fn_correct)

        if not os.path.isfile(correct_path):
            shutil.copyfile(output, correct_path)
            self.fail(
                msg="Output file {} missing so it was created.".format(fn_correct)
            )

        correct_hash = get_sha256(correct_path, **open_kwargs)
        self.assertEqual(
            correct_hash,
            output_hash,
            msg=(
                f"File contents hash mismatch: {correct_hash} != {output_hash}"
                f" comparing '{output}' with '{correct_path}'."
            )
            + self._debug_info(obj),
        )

    def assertFile(self, obj, field_path, fn, **kwargs):
        """Compare a process's output file to the given correct file.

        :param obj: object that includes the file to compare
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name

        :param str fn: file name (and relative path) of the correct
            file to compare against. Path should be relative to the
            ``tests/files`` directory of a Django application.

        :param str compression: if not ``None``, files will be
            uncompressed with the appropriate compression library
            before comparison.
            Currently supported compression formats are *gzip* and
            *zip*.

        :param filter: function for filtering the contents of output
            files. It is used in :func:`itertools.filterfalse` function
            and takes one parameter, a line of the output file. If it
            returns ``True``, the line is excluded from comparison of
            the two files.
        :type filter: ~types.FunctionType

        :param bool sort: if set to ``True``, basic sort will be performed
            on file contents before computing hash value.

        """
        field = dict_dot(obj.output, field_path)
        self._assert_file(obj, field["file"], fn, **kwargs)

    def assertFiles(self, obj, field_path, fn_list, **kwargs):
        """Compare a process's output file to the given correct file.

        :param obj: object which includes the files to compare
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            list of file names

        :param list fn_list: list of file names (and relative paths) of
            files to compare against. Paths should be relative to the
            ``tests/files`` directory of a Django application.

        :param str compression: if not ``None``, files will be
            uncompressed with the appropriate compression library
            before comparison.
            Currently supported compression formats are *gzip* and
            *zip*.

        :param filter: Function for filtering the contents of output
            files. It is used in :obj:`itertools.filterfalse` function
            and takes one parameter, a line of the output file. If it
            returns ``True``, the line is excluded from comparison of
            the two files.
        :type filter: ~types.FunctionType

        :param bool sort: if set to ``True``, basic sort will be performed
            on file contents before computing hash value.

        """
        field = dict_dot(obj.output, field_path)

        if len(field) != len(fn_list):
            self.fail(
                msg="Lengths of list:basic:file field and files list are not equal."
            )

        for fn_tested, fn_correct in zip(field, fn_list):
            self._assert_file(obj, fn_tested["file"], fn_correct, **kwargs)

    def assertFileExists(self, obj, field_path):
        """Ensure a file in the given object's field exists.

        :param obj: object that includes the file for which to check if
            it exists
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name/path
        """
        field = dict_dot(obj.output, field_path)
        output = obj.location.get_path(filename=field["file"])

        if not os.path.isfile(output):
            self.fail(msg="File {} does not exist.".format(field_path))

    def assertFilesExist(self, obj, field_path):
        """Ensure files in the given object's field exists.

        :param obj: object that includes list of files for which to check
            existance
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name/path
        """
        field = dict_dot(obj.output, field_path)

        for item in field:
            output_file = obj.location.get_path(filename=item["file"])
            if not os.path.isfile(output_file):
                self.fail(
                    msg="File {} in output field {} does not exist.".format(
                        item["file"], field_path
                    )
                )

    def assertDirExists(self, obj, field_path):
        """Assert that a directory in the output field of the given object exists.

        :param obj: object that includes the file for which to check if
            it exists

        :param field_path: directory name/path
        """
        schema, field = self._get_output_field(obj, field_path)
        if not schema["type"].startswith("basic:dir:"):
            self.fail(msg="Field {} is not of type basic:dir:".format(field_path))

        dir_path = obj.location.get_path(filename=field[field_path]["dir"])
        if not os.path.isdir(dir_path):
            self.fail(
                msg="Directory {} in output field {} does not exist.".format(
                    dir_path, field_path
                )
            )

    def _assert_dir_structure(self, dir_path, dir_struct, exact=True):
        """Compare tree structure of directory `dir_path` to `dir_struct`."""
        test_dirs = set()
        test_files = set()
        for root, dirs, files in os.walk(dir_path):
            for test_dir in dirs:
                test_dirs.add(os.path.relpath(os.path.join(root, test_dir), dir_path))
            for test_file in files:
                test_files.add(os.path.relpath(os.path.join(root, test_file), dir_path))

        def get_dirs(indict, root=""):
            """Generate directory names from dict."""
            for key, value in indict.items():
                if isinstance(value, dict):
                    yield os.path.join(root, key)
                    yield from get_dirs(value, os.path.join(root, key))
                elif value is not None:
                    self.fail(msg="Directory structure specification is incorrect")

        def get_files(indict, root=""):
            """Generate file names from dict."""
            for key, value in indict.items():
                if value is None:
                    yield os.path.join(root, key)
                elif isinstance(value, dict) and value:
                    yield from get_files(value, os.path.join(root, key))
                elif not isinstance(value, dict):
                    self.fail(msg="Directory structure specification is incorrect.")

        correct_files = {file for file in get_files(dir_struct)}
        correct_dirs = {correct_dir for correct_dir in get_dirs(dir_struct)}

        if exact and (test_dirs != correct_dirs or test_files != correct_files):
            self.fail(msg="Directory structure mismatch (exact check).")
        if not exact and (correct_dirs - test_dirs or correct_files - test_files):
            self.fail(msg="Directory structure mismatch (partial structure check).")

    def assertDirStructure(self, obj, field_path, dir_struct, exact=True):
        """Assert correct tree structure in output field of given object.

        Only names of directories and files are asserted. Content of files is
        not compared.

        :param obj: object that includes the directory to compare
        :type obj: ~resolwe.flow.models.Data

        :param str dir_path: path to the directory to compare

        :param dict dir_struct: correct tree structure of the directory.
            Dictionary keys are directory and file names with the correct nested
            structure. Dictionary value associated with each directory is a new
            dictionary which lists the content of the directory. Dictionary
            value associated with each file name is ``None``

        :param bool exact: if ``True`` tested directory structure must exactly
            match `dir_struct`. If ``False`` `dir_struct` must be a partial
            structure of the directory to compare

        """
        self.assertDirExists(obj, field_path)
        field = dict_dot(obj.output, field_path)
        dir_path = obj.location.get_path(filename=field["dir"])
        self._assert_dir_structure(dir_path, dir_struct, exact)

    def _assert_dir(self, dir_path, fn_correct, fail_on_funny=True):
        """Compare directory `dir_path` to compressed directory `fn_correct`."""
        correct_path = os.path.join(self.files_path, fn_correct)
        if not os.path.isfile(correct_path):
            with tarfile.open(correct_path, "w:gz") as f:
                for content in os.listdir(dir_path):
                    f.add(os.path.join(dir_path, content), arcname=content)
            self.fail(
                msg="Compressed output directory {} missing so it was created.".format(
                    fn_correct
                )
            )

        if not tarfile.is_tarfile(correct_path):
            self.fail(msg="{} is not a tar file.".format(fn_correct))

        with tempfile.TemporaryDirectory() as temp_dir:
            with tarfile.open(correct_path) as tar:
                tar.extractall(temp_dir)
            cmp = filecmp.dircmp(dir_path, temp_dir)
            if (
                cmp.left_only
                or cmp.right_only
                or cmp.diff_files
                or (fail_on_funny and cmp.funny_files)
            ):
                self.fail(
                    msg="Directory {} content mismatch: {}.".format(
                        fn_correct, cmp.report()
                    )
                )

    def assertDir(self, obj, field_path, fn):
        """Compare process output directory to correct compressed directory.

        :param obj: object that includes the directory to compare
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name

        :param str fn: file name (and relative path) of the correct compressed
            directory to compare against. Path should be relative to the
            ``tests/files`` directory of a Django application. Compressed
            directory needs to be in ``tar.gz`` format.

        """
        self.assertDirExists(obj, field_path)
        field = dict_dot(obj.output, field_path)
        dir_path = obj.location.get_path(filename=field["dir"])
        self._assert_dir(dir_path, fn)

    def assertJSON(self, obj, storage, field_path, file_name):
        """Compare JSON in Storage object to the given correct JSON.

        :param obj: object to which the
            :class:`~resolwe.flow.models.Storage` object belongs
        :type obj: ~resolwe.flow.models.Data

        :param storage: object or id which contains JSON to compare
        :type storage: :class:`~resolwe.flow.models.Storage` or
            :class:`str`

        :param str field_path: path to JSON subset in the
            :class:`~resolwe.flow.models.Storage`'s object to compare
            against. If it is empty, the entire object will be
            compared.

        :param str file_name: file name (and relative path) of the file
            with the correct JSON to compare against. Path should be
            relative to the ``tests/files`` directory of a Django
            application.

            .. note::

                The given JSON file should be compresed with *gzip* and
                have the ``.gz`` extension.

        """
        self.assertEqual(
            os.path.splitext(file_name)[1], ".gz", msg="File extension must be .gz"
        )

        if not isinstance(storage, Storage):
            storage = Storage.objects.get(pk=storage)

        storage_obj = dict_dot(storage.json, field_path)

        file_path = os.path.join(self.files_path, file_name)
        if not os.path.isfile(file_path):
            with gzip.open(file_path, mode="wt") as f:
                json.dump(storage_obj, f)

            self.fail(msg="Output file {} missing so it was created.".format(file_name))

        with gzip.open(file_path, mode="rt") as f:
            file_obj = json.load(f)

        self.assertAlmostEqualGeneric(
            storage_obj,
            file_obj,
            msg="Storage {} field '{}' does not match file {}".format(
                storage.id, field_path, file_name
            )
            + self._debug_info(obj),
        )

    def _debug_info(self, data):
        """Return data's debugging information."""
        msg_header = "Debugging information for data object {}".format(data.pk)
        msg = (
            "\n\n"
            + len(msg_header) * "="
            + "\n"
            + msg_header
            + "\n"
            + len(msg_header) * "="
            + "\n"
        )
        path = data.location.get_path(filename="stdout.txt") if data.location else None
        if path and os.path.isfile(path):
            msg += "\nstdout.txt:\n" + 11 * "-" + "\n"
            with io.open(path, mode="rt") as fn:
                msg += fn.read()

        if data.process_error:
            msg += "\nProcess' errors:\n" + 16 * "-" + "\n"
            msg += "\n".join(data.process_error)

        if data.process_warning:
            msg += "\nProcess' warnings:\n" + 18 * "-" + "\n"
            msg += "\n".join(data.process_warning)

        return msg



"""Compute hashes from stream."""

import hashlib
from io import RawIOBase
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

import crcmod

if TYPE_CHECKING:
    from os import PathLike


def compute_hashes(file: "PathLike[str]") -> Dict[str, str]:
    """Compute hashes for a given file/directory.

    :param file_: path-like object pointing to a file/directory.

    :returns: dictionary that contains hash types as keys and corresponding
        hashes as values. There is one entry in this dictionary for each hash
        type in StreamHasher.KNOWN_HASH_TYPES.
        If file_ points to a directory values are empty strings.
    """
    path = Path(file)
    if path.is_dir():
        return {hash_type: "" for hash_type in StreamHasher.KNOWN_HASH_TYPES}

    hasher = StreamHasher()
    with path.open("rb") as stream:
        hasher.compute(stream)
    return {
        hash_type: hasher.hexdigest(hash_type)
        for hash_type in StreamHasher.KNOWN_HASH_TYPES
    }


class AWSS3ETagHash:
    """Compute AWSS3etag hash.

    See https://teppen.io/2018/10/23/aws_s3_verify_etags/ for details.
    """

    def __init__(self):
        """Initialize hasher."""
        self.md5_digests = []
        self.hasher = hashlib.md5

    def update(self, data: bytes):
        """Update hasher with given data."""
        self.md5_digests.append(self.hasher(data).digest())

    def hexdigest(self) -> str:
        """Compute and return hexdigest."""
        if len(self.md5_digests) == 1:
            return self.md5_digests[0].hex()
        else:
            joined_md5 = b"".join(self.md5_digests)
            return "{}-{}".format(
                self.hasher(joined_md5).hexdigest(), len(self.md5_digests)
            )


class StreamHasher:
    """Compute hash for data in the stream."""

    KNOWN_HASH_TYPES = ["md5", "crc32c", "awss3etag"]
    _hashers = {
        "awss3etag": AWSS3ETagHash,
        "md5": hashlib.md5,
        "crc32c": lambda: crcmod.predefined.PredefinedCrc("crc32c"),
    }

    def __init__(self, hashes: Optional[List[str]] = None, chunk_size=8 * 1024 * 1024):
        """Initialize the hasher class using given chunk_size.

        Optionally set the hashes we want to compute by setting the hashes
        property. The hashes property must include strings from
        KNOWN_HASH_TYPES list.

        Be careful to set the correct chunk_size for AWSS3ETag computation.
        It must be the same as upload chunk size used to upload file to S3.
        """
        self.chunk_size = chunk_size
        self.hashes = hashes or StreamHasher.KNOWN_HASH_TYPES

    def _init_hashers(self):
        """Initialize known hashers."""
        self._hashers = {name: StreamHasher._hashers[name]() for name in self.hashes}

    def compute(self, stream_in: RawIOBase, stream_out: RawIOBase = None):
        """Compute and return the hash for the given stream.

        The data is read from stream_in until EOF, given to the hasher, and
        written unmodified to the stream_out (unless it is None).

        :param stream_in: input stream.
        :type stream_in: io.RawIOBase

        :param stream_out: output stream.
        :type stream_out: io.RawIOBase
        """
        self._init_hashers()
        read_bytes = self.chunk_size
        while read_bytes == self.chunk_size:
            data = stream_in.read(self.chunk_size)
            read_bytes = len(data)
            for hasher in self._hashers.values():
                hasher.update(data)
            if stream_out is not None:
                stream_out.write(data)

    def digest(self, hash_type: str) -> bytes:
        """Return the digest for the given hash_type.

        :raises KeyError: when hash_type is unknown. Known hash types are
            listed in Hasher.KNOWN_HASH_TYPES.
        :return: computed hash digest of the given type.
        :rtype: bytes
        """
        return bytes.fromhex(self.hexdigest(hash_type))

    def hexdigest(self, hash_type: str) -> str:
        """Return the hex digest for the given hash_type.

        :raises KeyError: when hash_type is unknown. Known hash types are
            listed in Hasher.KNOWN_HASH_TYPES.
        :return: computed hash hexdigest of the given type.
        :rtype: str
        """
        return self._hashers[hash_type].hexdigest().lower()



"""Collection viewset."""

from django.db import transaction
from django.db.models import F, Func, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, mixins, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.flow.filters import CollectionFilter
from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity
from resolwe.flow.serializers import CollectionSerializer
from resolwe.flow.serializers.annotations import AnnotationFieldDictSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import PermissionModel

from .mixins import (
    ResolweBackgroundDeleteMixin,
    ResolweBackgroundDuplicateMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)


class BaseCollectionViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    ResolweBackgroundDeleteMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ResolweBackgroundDuplicateMixin,
    viewsets.GenericViewSet,
):
    """Base API view for :class:`Collection` objects."""

    qs_permission_model = PermissionModel.objects.select_related("user", "group")

    # Add entity count as a subquery: by default Django joins tables and this explodes
    # when one tries to annotate the collection query with data and entity count at the
    # same time.
    entity_count_subquery = (
        Entity.objects.filter(collection_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_count_subquery = (
        Data.objects.filter(collection_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(collection_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )

    qs_descriptor_schema = DescriptorSchema.objects.select_related("contributor")

    queryset = (
        Collection.objects.select_related("contributor")
        .prefetch_related(Prefetch("descriptor_schema", queryset=qs_descriptor_schema))
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
        .annotate(entity_count=Subquery(entity_count_subquery))
    )

    filterset_class = CollectionFilter
    permission_classes = (get_permissions_class(),)

    ordering_fields = (
        "contributor",
        "contributor__first_name",
        "contributor__last_name",
        "collection__name",
        "created",
        "id",
        "modified",
        "name",
    )
    ordering = "id"

    def get_queryset(self):
        """Prefetch permissions for current user."""
        return self.prefetch_current_user_permissions(self.queryset)

    def create(self, request, *args, **kwargs):
        """Only authenticated users can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)


class CollectionViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for :class:`Collection` objects."""

    serializer_class = CollectionSerializer

    @extend_schema(
        request=AnnotationFieldDictSerializer(),
        responses={status.HTTP_204_NO_CONTENT: None},
    )
    @action(detail=True, methods=["post"])
    @transaction.atomic
    def set_annotation_fields(self, request, pk=None):
        """Set AnnotationFields on collection."""
        # No need to check for permissions, since post requires edit by default.
        collection = self.get_object()
        # Read and validate the request data.
        serializer = AnnotationFieldDictSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        # Set the new annotation fields.
        collection.annotation_fields.set(serializer.validated_data["annotation_fields"])
        return Response(status=status.HTTP_204_NO_CONTENT)
